from unipressed import UniprotkbClient
import pandas as pd
import os
from pathlib import Path
import glob
import json
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from textwrap import wrap
from pymol import cmd
import cv2 as cv
import sys

import alphascreen
from alphascreen import argparser
from alphascreen import jobsetup
from alphascreen import analyze

def decide():

    ##################################
    """
    This is the master function that is used to determine what the user is asking for from the command-line arguments,
    and set up the variables required to call those functions
    """
    ##################################

    #Get the passed parameters
    params = argparser.argparse()

    table = params['table']
    focus = params['focus']
    fraglen = params['fraglen']
    overlap = params['overlap']
    dimerize = params['dimerize']
    dimerize_all = params['dimerize_all']
    dimerize_except = params['dimerize_except']
    consider = params['consider']
    dontwrite = params['dontwrite']
    alphafold_exec = params['alphafold_exec']
    columnA = params['columnA']
    columnB = params['columnB']
    check = params['check']
    check_write = params['check_write']
    threshold = params['threshold']
    overwrite = params['overwrite']
    writetable = params['writetable']
    rankby = params['rankby']
    exhaustive = params['exhaustive']
    showall = params['showall']

    if fraglen == "":
        fraguniprot = 0
        fraglen = 500
    elif len(fraglen.split("/")) == 2:
        if len(table.split("/")) != 2:
            sys.exit("\n>> Error: You provided two fragment lengths but the input was not two uniprot IDs.\n")
        fraguniprot = [table.split("/")[0], table.split("/")[1]]
        fraglen = [int(fraglen.split("/")[0]), int(fraglen.split("/")[1])]
    else:
        fraguniprot = 0
        fraglen = int(fraglen)

    if table != "" and not dontwrite:

        with open("log.txt", 'w') as f:
            f.write("AlphaScreen version: " + str(alphascreen.__version__) + "\n")
            f.write("Input: " + table + "\n")
            f.write("Fragment: " + str(fraglen) + "\n")
            f.write("Overlap: " + str(overlap) + "\n")
            if dimerize:
                f.write("Dimerized: " + dimerize + "\n")
            elif dimerize_all:
                f.write("Dimerized all proteins\n")
            elif dimerize_except:
                f.write("Dimerize all except: " + dimerize_except + "\n")
            if consider != "":
                f.write("Consider: " + consider + "\n")
            f.write("Alphafold executable: " + alphafold_exec + "\n")
            if exhaustive:
                f.write("Exhaustive = True.\n")


    ##################################
    #Parse input

    towrite = not dontwrite

    if check:
        towrite = False
        jobsetup.findunfinished(alphafold_exec, write=towrite)
        sys.exit()
    elif check_write:
        towrite = True
        jobsetup.findunfinished(alphafold_exec, write=towrite)
        sys.exit()

    if rankby=="pae":
        rankby="scaledPAE"

    if showall:
        threshold = 0

    ##################################

    if writetable:
        print("\n>> Parsing results...")
        df = analyze.getscores(rankby)
        analyze.write_top(df, 0, rankby)
        sys.exit()

    ##################################

    if threshold != -1:
        print("\n>> Parsing results...")
        df = analyze.getscores(rankby)
        
        if df.empty:
            sys.exit("\n>> Error: no results could be found.\n")
        try:
            df[df[rankby]>threshold]
        except TypeError:
            sys.exit("\n>> Error: some values you want to rank by don't seem to exist. Did you try to rank by iptm without those scores existing?\n")

        if df[df[rankby]>threshold].empty:
            sys.exit("\n>> Error: no results could be found with "+rankby+" above " + str(threshold) + ".\n")

        analyze.summarize_pae_pdf(df, threshold, rankby)
        analyze.write_top(df, threshold, rankby)
        analyze.write_modelpngs(df, threshold, rankby, overwrite=overwrite)
        analyze.summarize_paeandmodel_pdf(df, threshold, rankby)
        sys.exit()

    ##################################

    consideruniprot, considerstart, considerend = [], [], []
    if consider != "":
        if consider[-4:] == ".txt":
            with open(consider) as f:
                considerlist=[line.strip() for line in f.readlines()]
            for c in considerlist:
                considerargs = c.split("/")
                if len(considerargs) != 3:
                    sys.exit("\n>> Error: there is a problem with the line " + c + " in " + consider + ".\n")
                consideruniprot.append(considerargs[0])
                considerstart.append(int(considerargs[1])-1)
                considerend.append(int(considerargs[2])-1)
        else:
            considerargs = consider.split("/")
            if len(considerargs) != 3:
                sys.exit("\n>> Error: the --consider argument was not passed properly.\n")
            consideruniprot = [considerargs[0]]
            considerstart = [int(considerargs[1])-1]
            considerend = [int(considerargs[2])-1]

    filetype=""
    if table != "":
        if table[-4:] == ".txt":
            filetype = "table"
        elif table[-4:] == "xlsx":
            filetype = "excel"

        if filetype!="":
            Ainteractors, Binteractors = jobsetup.getinteractors(table, filetype, columnA, columnB, focus, exhaustive)
        else: #if two proteins
            Ainteractors = [table.split("/")[0]]
            Binteractors = [table.split("/")[1]]
            print("\n>> Parsing " + Ainteractors[0] + " and " + Binteractors[0] + "...\n")

        jobsetup.getfastas_writecommands(Ainteractors, Binteractors, consideruniprot, considerstart, considerend, split=True,fraguniprot=fraguniprot,fraglen=fraglen,overlap=overlap,dimerize=dimerize,dimerize_all=dimerize_all,dimerize_except=dimerize_except,write=towrite,alphafold_exec=alphafold_exec)

    ##################################




import optparse
import sys
import alphascreen

def argparse():

    """
    optparse is used to initialize all command-line options.
    """
    
    parser = optparse.OptionParser(usage="Usage: %prog [options]",
        version=alphascreen.__version__)

    jobsetup_opts = optparse.OptionGroup(
        parser, 'Job setup')

    jobsetup_opts.add_option("--parse",
        action="store", dest="table", type="string", default="", metavar='file',
        help="Path to the excel file (.xlsx) or table (.txt).")

    jobsetup_opts.add_option("--fragment",
        action="store", dest="fraglen", type="string", default="", metavar='fragment-length',
        help="Approximate fragment length. Default is 500. For shorter fragments, try 250.")

    jobsetup_opts.add_option("--overlap",
        action="store", dest="overlap", type="int", default=50, metavar='overlap-length',
        help="Sequence is extended by this amount on either side of slices. Default is 50.")

    jobsetup_opts.add_option("--exhaustive",
        action="store_true", dest="exhaustive", default=False,
        help="Run every protein in column A against every protein in column B.")

    jobsetup_opts.add_option("--dimerize",
        action="store", dest="dimerize", type="string", default="", metavar='dimerize-id',
        help="Uniprot ID to dimerize. Alternatively, provide a text file (.txt) with a single column list of uniprot IDs to dimerize.")

    jobsetup_opts.add_option("--dimerize_all",
        action="store_true", dest="dimerize_all", default=False,
        help="Dimerize all proteins.")

    jobsetup_opts.add_option("--dimerize_all_except",
        action="store", dest="dimerize_except", type="string", default="", metavar='ids-not-to-dimerize',
        help="Dimerize all proteins except the uniprot ID provided here. Alternatively, provide a text file (.txt) with a single column list of uniprot IDs to NOT dimerize. Everything else will be dimerized.")

    jobsetup_opts.add_option("--consider",
        action="store", dest="consider", type="string", default="", metavar='sequence-to-consider',
        help="Uniprot ID and sequence range to consider. Example: \"Q86VS8/1/200\" only considers amino acids 1-200 for uniprot ID Q86VS8. Alternatively, provide a text file with a single column list of sequences to consider with a similar syntax (id/start/end).")

    jobsetup_opts.add_option("--alphafold_exec",
        action="store", dest="alphafold_exec", type="string", default="colabfold2", metavar='executable',
        help="Colabfold executable. Default is \"colabfold2\"")

    jobsetup_opts.add_option("--columnA",
        action="store", dest="columnA", type="string", default="SWISS-PROT Accessions Interactor A", metavar='columnA-name',
        help="Name of column heading for uniprot IDs for first interactors.")

    jobsetup_opts.add_option("--columnB",
        action="store", dest="columnB", type="string", default="SWISS-PROT Accessions Interactor B", metavar='columnB-name',
        help="Name of column heading for uniprot IDs for second interactors.")

    jobsetup_opts.add_option("--focus",
        action="store", dest="focus", type="string", default="", metavar='uniprot-id',
        help="Uniprot ID to focus on. This means that it will the first chain in any predictions that contain it.")

    jobsetup_opts.add_option("--dontwrite",
        action="store_true", dest="dontwrite", default=False,
        help="Pass if you don't want to write out files.")
    
    parser.add_option_group(jobsetup_opts)

    check_opts = optparse.OptionGroup(
        parser, 'Check runs')

    check_opts.add_option("--check",
        action="store_true", dest="check", default=False,
        help="Check how many jobs are left.")

    check_opts.add_option("--write_unfinished",
        action="store_true", dest="check_write", default=False,
        help="Check how many jobs are left and write a new script to only run those jobs.")

    parser.add_option_group(check_opts)

    analyze_opts = optparse.OptionGroup(
        parser, 'Analysis')

    analyze_opts.add_option("--show_top",
        action="store", dest="threshold", type="float", default=-1, metavar="threshold",
        help="Summarize results for top hits (above the threshold specified here) in tables and pdfs.")

    analyze_opts.add_option("--show_all",
        action="store_true", dest="showall", default=False,
        help="Summarize results for all hits in tables and pdfs.")

    analyze_opts.add_option("--write_table",
        action="store_true", dest="writetable", default=False,
        help="Write the full results as a table ranked by iptm.")

    jobsetup_opts.add_option("--rankby",
        action="store", dest="rankby", type="string", default="pae", metavar='ptm-iptm-pae',
        help="Score by which models are ranked (pae, ptm, or iptm). Default is pae. This is used for both choosing the best model in a prediction and then ranking the best models in the summaries.")
    
    analyze_opts.add_option("--overwrite",
        action="store_true", dest="overwrite", default=False,
        help="Overwrite snapshot pngs if they already exist. Default is False.")

    parser.add_option_group(analyze_opts)

    """
    The rest of the function parses the input and generates a dictionary for use in decisiontree.py
    """

    #Get the passed arguments from command-line
    options,args = parser.parse_args()

    #If there are less than 1 arguments passed, there were no options passed.
    if len(sys.argv) < 1:
            #parser.print_help()
            #print("\n>> You did not pass the arguments properly. For help, run \"starparser -h\".")
            print("\n>> Help: alphascreen -h\n")
            sys.exit()

    #Initialize an empty dictionary to place all the parameters in
    params={}

    #Place the passed parameters (or default values if none were passed) into the params dictionary
    for i in options.__dict__.items():
        params[i[0]] = i[1]
        
    #The dictionary is the main input to decisiontree.py
    return(params)


from unipressed import UniprotkbClient
import pandas as pd
import os
from pathlib import Path
import sys
import numpy as np

def getuni(ACCESSION):
    ENTRY = UniprotkbClient.fetch_one(ACCESSION)
    SEQ = ENTRY['sequence']['value']
    NAME = ENTRY['uniProtkbId']
    return(NAME, SEQ)

def getinteractors(file,filetype, columnA, columnB, focus, exhaustive):
    
    print("\n>> Reading table...\n")
    if filetype == "table":
        table = pd.read_table(file, usecols=[columnA, columnB])
    elif filetype == "excel":
        table = pd.read_excel(file, usecols=[columnA, columnB])
    
    Ainteractors = table[columnA].tolist()
    Binteractors = table[columnB].tolist()

    #cleanup
    if exhaustive:
        Ainteractors = [x for x in Ainteractors if x == x] #remove nan float which result from empty cells
        Binteractors = [x for x in Binteractors if x == x]
    else:
        Ainteractors = ["empty-cell" if x != x else x for x in Ainteractors] #replace nan-float with 0, which result from empty cells
        Binteractors = ["empty-cell" if x != x else x for x in Binteractors]
    Ainteractors = [s.strip() for s in Ainteractors]
    Binteractors = [s.strip() for s in Binteractors]

    if focus != "":
        if focus not in Ainteractors or focus not in Binteractors:
            sys.exit("!! Error: the uniprot ID " + focus + " to focus on was not found.\n")
        Ainteractors_fixed = []
        Binteractors_fixed = []
        for A, B in zip(Ainteractors,Binteractors):
            if B == focus:
                Ainteractors_fixed.append(B)
                Binteractors_fixed.append(A)
            else:
                Ainteractors_fixed.append(A)
                Binteractors_fixed.append(B)
        Ainteractors = Ainteractors_fixed
        Binteractors = Binteractors_fixed

    if exhaustive:
        Ainteractors = list(set(Ainteractors))
        Binteractors = list(set(Binteractors))
        Ainteractors_exh = []
        Binteractors_exh = []
        for A in Ainteractors:
            for B in Binteractors:
                Ainteractors_exh.append(A)
                Binteractors_exh.append(B)
        Ainteractors = Ainteractors_exh
        Binteractors = Binteractors_exh

    return(Ainteractors, Binteractors)

def getfastas_writecommands(Ainteractors, Binteractors, consideruniprot,considerstart,considerend, split=True,fraguniprot=0,fraglen=500,overlap=50,dimerize="",dimerize_all=False,dimerize_except="",write=True,alphafold_exec="colabfold2"):
    
    dimerizelst = []
    dontdimerizelst = []

    if dimerize_all:
        print(">> Dimerizing all proteins...\n")

    elif dimerize_except:
        if dimerize_except[-4:] == ".txt":
            print(">> Dimerizing all proteins except those in " + dimerize_except + "...\n")
            with open(dimerize_except) as f:
                dontdimerizelst=[line.strip() for line in f.readlines()]
        else:
            dontdimerzelst = [dimerize_except]
            print(">> Dimerizing all proteins except " + dimerize_except + "...\n")

    elif dimerize != "":
        if dimerize[-4:] == ".txt":
            print(">> Dimerizing uniprot IDs in " + dimerize + "...\n")
            with open(dimerize) as f:
                dimerizelst=[line.strip() for line in f.readlines()]
        else:
            dimerizelst = [dimerize]
            print(">> Dimerizing uniprot ID " + dimerize + "...\n")

    #remove duplicates
    repeating_indices = []
    for i, pair in enumerate(zip(Ainteractors, Binteractors)):
        for j, pair_same in enumerate(zip(Ainteractors, Binteractors)):
            if pair[0] == pair_same[0] and pair[1] == pair_same[1] and i != j: #i != j means that one instance is retained
                repeating_indices.append(j)
                #print(pair[0], pair[1], pair_same[0], pair_same[1])
            if pair[0] == pair_same[1] and pair[1] == pair_same[0]:
                repeating_indices.append(j)
                #print(pair[0], pair[1], pair_same[0], pair_same[1])
    Ainteractors = [j for i, j in enumerate(Ainteractors) if i not in repeating_indices]
    Binteractors = [j for i, j in enumerate(Binteractors) if i not in repeating_indices]
    
    if len(Ainteractors) > 1 and len(Binteractors) > 1:
        print(">> Removed " + str(len(repeating_indices)) + " duplicates...\n")
    
    os.makedirs('fastas', exist_ok=True)
    os.makedirs('results', exist_ok=True)

    for fname in os.listdir('fastas'):
        if fname.endswith('.fasta'):
            sys.exit(">> Error: it looks like fasta files already exist in the fastas folder.\n")

    colabfoldcommand = []
    unilst = []
    failed = []

    if len(dontdimerizelst) != 0:
        alluniprots = list(set(list(Ainteractors+Binteractors)))
        dimerizelst = [i for i in alluniprots if i not in dontdimerizelst]

    print(">> Getting sequences...\n")

    Unis_found = []
    Names_found = []
    Seqs_found = []

    notconsidered = consideruniprot.copy() #fully populated and will be removed one by one to check that all were found
    for A, B in zip(Ainteractors, Binteractors):

        try:
            if A in Unis_found:
                Aindex = Unis_found.index(A)
                Aname = Names_found[Aindex]
                Aseq = Seqs_found[Aindex]
            else:
                Aname,Aseq=getuni(A)
                Unis_found.append(A)
                Names_found.append(Aname)
                Seqs_found.append(Aseq)
        except:
            failed.append(A)
            continue
        try:
            if B in Unis_found:
                Bindex = Unis_found.index(B)
                Bname = Names_found[Bindex]
                Bseq = Seqs_found[Bindex]
            else:
                Bname,Bseq=getuni(B)
                Unis_found.append(B)
                Names_found.append(Bname)
                Seqs_found.append(Bseq)
        except:
            failed.append(B)
            continue
            
        unilst.append(A + ":" + Aname)
        unilst.append(B + ":" + Bname)

        addmeA=0
        addmeB=0
        if len(consideruniprot) > 0:

            if A in consideruniprot:

                a = consideruniprot.index(A)
                if considerend[a] > len(Aseq):
                    sys.exit(">> Error: last amino acid to consider is outside the protein length for " + consideruniprot[a] + ". Maximum is " + str(len(Aseq)+1) + ".\n")
                Aseq = Aseq[considerstart[a]:considerend[a]]
                addmeA = considerstart[a]

                if A in notconsidered:
                    notconsidered.remove(A) 

            if B in consideruniprot:

                b = consideruniprot.index(B)
                if considerend[b] > len(Bseq):
                    sys.exit(">> Error: last amino acid to consider is outside the protein length for " + consideruniprot[b] + ". Maximum is " + str(len(Bseq)+1) + ".\n")
                Bseq = Bseq[considerstart[b]:considerend[b]]
                addmeB = considerstart[b]

                if B in notconsidered:
                    notconsidered.remove(B) 
        
        if split:

            if type(fraglen) == list:
                if A in fraguniprot:
                    fraglen_cur=fraglen[fraguniprot.index(A)]
                else:
                    fraglen_cur = 500
            else:
                fraglen_cur = fraglen

            Anamelst, Aseqlst = splitfasta(Aname,Aseq,fraglen_cur,overlap,addmeA)

            if type(fraglen) == list:
                if B in fraguniprot:
                    fraglen_cur=fraglen[fraguniprot.index(B)]
                else:
                    fraglen_cur = 500
            else:
                fraglen_cur = fraglen

            Bnamelst, Bseqlst = splitfasta(Bname,Bseq,fraglen_cur,overlap,addmeB)

        else:

            Anamelst = [Aname]
            Aseqlst = [Aseq]
            Bnamelst = [Bname]
            Bseqlst = [Bseq]
            
        
        for An, As in zip(Anamelst, Aseqlst):
            
            for Bn, Bs in zip(Bnamelst, Bseqlst):

                fastaname = 'fastas/'+An+'-'+Bn+'.fasta'
                fastaname_minimal = 'fastas/'+An+'-'+Bn+'.fasta'
                
                if write:
                    
                    with open(fastaname, 'w') as f:
                        
                        f.write(">"+An+"\n"+As+"\n")
                        
                        if A in dimerizelst or dimerize_all:
                            f.write(">"+An+"\n"+As+"\n")
                            
                        f.write(">"+Bn+"\n"+Bs+"\n")
                        
                        if B in dimerizelst or dimerize_all:
                            f.write(">"+Bn+"\n"+Bs+"\n")

                colabfoldcommand.append(alphafold_exec + " " + fastaname_minimal + " --output=results/" + An + "-" + Bn)

    if len(failed) > 0:
        failed = list(set(failed))
        print("!! Warning: there was a problem getting the Uniprot data for accessions:\n")
        for f in failed:
            print("·· "+ str(f) + "\n")

    if len(notconsidered) > 0:
        for c in notconsidered:
            print("!! Warning: you asked to consider a region within " + c + ", but it wasn't in your input.\n")

    colabfoldcommand=list(dict.fromkeys(colabfoldcommand)) 

    if write:
        print(">> Writing " + str(len(colabfoldcommand)) + " Alphafold commands...\n")
        with open("runpredictions.bsh", 'w') as f:
            for c in colabfoldcommand:
                f.write(c+"\n")
                
        with open("uniprots.txt", 'w') as f:
            for u in list(set(unilst)):
                f.write(u+"\n")

        print(">> Run the Alphafold jobs with \"bash runpredictions.bsh\"\n")
    else:
        print(">> There are " + str(len(colabfoldcommand)) + " Alphafold commands...\n")
            
    print("Done!\n")

def splitfasta(name,seq,fraglen,overlap,addme):

    seqlen = len(seq)
    #print(name, seqlen)
    
    splitinto = round(seqlen/fraglen)
    #print(splitinto)
    
    test=[]
    if splitinto in [0, 1]:
        index_withOverlap = [[0,seqlen]]
    else:
        currfraglen = int(seqlen/splitinto)
        middlefragindices=[]
        for n in range(1,splitinto):
            if n == 1 and splitinto > 2:
                middlefragindices.append(currfraglen*n+int(overlap/2)) #account for overlap
            elif n == splitinto-1 and splitinto > 2:
                middlefragindices.append(currfraglen*n-int(overlap/2)) #account for overlap
            else:
                middlefragindices.append(currfraglen*n) 
        
        #print(middlefragindices)
    
        index_withOverlap=[]
        for i,m in enumerate(middlefragindices):
            if i == 0:
                index_withOverlap.append([0, m+overlap])
            else:
                index_withOverlap.append([middlefragindices[i-1]-overlap, m+overlap])
                
        index_withOverlap.append([middlefragindices[-1]-overlap, seqlen])
                    
        #print(index_withOverlap)
        
    splitlens_withOverlap = []
    for i in index_withOverlap:
        splitlens_withOverlap.append(i[1]-i[0])
    #print("New lengths", splitlens_withOverlap)
        
    splitseqs_withOverlap = []
    newnames = []
    
    for i in index_withOverlap:
        
        splitseqs_withOverlap.append(seq[i[0]:i[1]])
        newnames.append(name+"_"+str(i[0]+1+addme)+"_"+str(i[1]+1+addme))
        
    #print(newnames)
        
    return(newnames, splitseqs_withOverlap)

def findunfinished(alphafold_exec, write=True):
    
    colabfoldcommand = []
    numfound=0

    warned=False    
    for fastapath in Path('fastas/').rglob('**/*.fasta'):
        
        resultdir = "results" + str(fastapath).split("fastas")[1][:-6]

        if not os.path.exists(resultdir) and not warned:
            print("\n!! Warning: could not find one of the directories. Make sure this the right directory where the Alphafold jobs were run.")
            warned=True
        
        found = False
        if os.path.exists(resultdir):
            for fname in os.listdir(resultdir):
                if fname.endswith('.pdb'):
                    found = True
                    numfound+=1
                    break
        if not found:
            colabfoldcommand.append(alphafold_exec + " fastas/" + str(fastapath).split("fastas/")[1] + " --output=results" + str(resultdir).split("results")[1] + "\n")

                
    print("\n>> Finished runs: " + str(numfound) + "\n>> Unfinished runs: " + str(len(colabfoldcommand))+"\n")
    
    colabfoldcommand=list(dict.fromkeys(colabfoldcommand)) 

    if write:
        if len(colabfoldcommand) == 0:
            sys.exit("\n>> There are no jobs left.\n")
        with open("runpredictions-unfinished.bsh", 'w') as f:
            for c in colabfoldcommand:
                f.write(c)
        print(">> Wrote runpredictions-unfinished.bsh\n")



from distutils.core import setup
from setuptools import find_packages
import alphascreen

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='alphascreen',
      version=alphascreen.__version__,
      description='Use colabfold to screen protein-protein interactions.',
      author='Sami Chaaban',
      author_email='chaaban@mrc-lmb.cam.ac.uk',
      url='http://pypi.python.org/pypi/alphascreen/',
      #long_description=long_description,
      #long_description_content_type='text/markdown',
      packages=find_packages(),
      entry_points={
          "console_scripts": [
            "alphascreen = alphascreen.__main__:main",
            ],
      },
      install_requires=["pandas","unipressed", "matplotlib", "numpy", "opencv-python", "openpyxl"],
      python_requires='>=3.8'
     )



import json
import matplotlib.pyplot as plt
from pathlib import Path
import numpy as np
import pandas as pd
import os
from matplotlib.backends.backend_pdf import PdfPages
from textwrap import wrap
from pymol import cmd
import cv2 as cv
import glob
import sys
import time
import pymol
import math

def getscores(rankby):

    proteinA_lst=[]
    proteinB_lst=[]
    uniprotA_lst=[]
    uniprotB_lst=[]
    iptmhighscore_lst = []
    ptmhighscore_lst = []
    mincrosspae_lst = []
    inv_mincrosspae_lst = []
    pdbname_lst = []
    paepng_lst = []
    paejson_lst = []
    
    existuniprot = False
    #get uniprots
    if os.path.exists("uniprots.txt"):
        existuniprot = True
    uniprotlst = []
    if existuniprot:
        with open("uniprots.txt") as f:
            lines=f.readlines()
        for l in lines:
            uniprotlst.append([l.split(":")[0], l.split(":")[1].replace("\n","")])
        #print(uniprotlst)

    skipped=0

    for resultdir in Path('results').glob('*/'):

        resultdir = str(resultdir)
        resultname = resultdir.split("/")[-1]
            
        try:
            next(Path(resultdir).glob("*rank*.pdb"))
        except StopIteration:
            skipped+=1
            #print("\n>> Warning: skipping " + resultname + " since it is missing PDB files.")
            continue
            
        try:
            next(Path(resultdir).glob("*_scores.json"))
            scorewildcard="_scores.json"
        except StopIteration:
            try:
                next(Path(resultdir).glob("*_pae.json"))
                scorewildcard="_pae.json"
            except StopIteration:
                print("\n>> Warning: skipping " + resultname + " since it is missing the scores json files.")
                skipped+=1
                continue

        modelnumlst = []
        pdbpaths = []
        for path in Path(resultdir).glob("*rank*.pdb"):
            pdbpaths.append(str(path))
            modelnum = str(path).split("model_")[-1].split(".pdb")[0].split("_")[0]
            modelnumlst.append(int(modelnum)-1) #-1 is to turn it into index (0-4) since numbers are 1-5

        if os.path.exists(resultdir+"/scores.txt"):
            with open(resultdir+"/scores.txt") as f:
                lines=f.readlines()
                iptms=[float(x.split(' ')[0].split(":")[1]) for x in lines]
                ptms=[float(x.split(' ')[1].split(":")[1]) for x in lines]
        else:
            iptms=["-"] * len(pdbpaths)
            ptms=["-"] * len(pdbpaths)
            if rankby in ["iptm, ptm"]:
                print("\n>> Warning: skipping " + resultname + " since it is missing the scores.txt file with the iptms and ptms.")
                skipped+=1
                continue

        try:
            next(Path(resultdir).glob("*PAE*.png"))
            for path in Path(resultdir).glob("*PAE*.png"):
                paepng_lst.append(str(path))
                break
        except StopIteration:
            paepng_lst.append("-")
        
        ProteinA=resultname.split("-")[0]
        ProteinB=resultname.split("-")[1]
        ProteinAmin=ProteinA.split("_")[0]+"_"+ProteinA.split("_")[1]
        ProteinBmin=ProteinB.split("_")[0]+"_"+ProteinB.split("_")[1]
        proteinA_lst.append(ProteinA)
        proteinB_lst.append(ProteinB)
         
        paejsons=[]
        paenumlst=[]
        if scorewildcard == "_scores.json":
            for path in Path(resultdir).glob("*_model_*_scores.json"):
                paejsons.append(str(path))
                paenumlst.append(int(str(path).split("_scores.json")[0].split("_model_")[-1])-1)
        elif scorewildcard == "_pae.json":
            for path in Path(resultdir).glob("rank_*_model_*_ptm_seed_0_pae.json"):
                paejsons.append(str(path))
                paenumlst.append(int(str(path).split("_ptm_seed_0_pae.json")[0].split("_model_")[-1])-1)
        
        if existuniprot:
            foundA=False
            foundB=False
            for u in uniprotlst:
                if u[1] == ProteinAmin and not foundA:
                    uniprotA_lst.append(u[0])
                    foundA=True
                if u[1] == ProteinBmin and not foundB:
                    uniprotB_lst.append(u[0])
                    foundB=True
                if foundA and foundB:
                    break
            if not foundA:
                print("Could not find uniprot ID for " + ProteinAmin)
            if not foundB:
                print("Could not find uniprot ID for " + ProteinBmin)


        if rankby in ["iptm", "ptm"]:
            if rankby == "iptm":
                m = np.argmax(iptms)
            else:
                m = np.argmax(ptms)
        elif rankby=="scaledPAE":
            minpaes=[]
            for paejson in paejsons:
                minpaes.append(getmincrosspae(paejson))
            m = paenumlst[np.argmin(minpaes)]
            # For the future: if same crosspae, choose best iptm
            # if len([k for k in minpaes if k == min(minpaes)]) > 1:
            #     print("there are multiple solutions for " + str(scorepath))
        else:
            sys.exit("\n>> Provide ptm, iptm, or pae as the ranking method.\n")

        iptmhighscore_lst.append(iptms[m])
        ptmhighscore_lst.append(ptms[m])

        paejson_lst.append(paejsons[paenumlst.index(m)])
        if rankby=="scaledPAE":
            mincrosspae_lst.append(minpaes[paenumlst.index(m)])
            inv_mincrosspae_lst.append(1-(minpaes[paenumlst.index(m)]/30))
        else:
            mincrosspae_lst.append("-")
            inv_mincrosspae_lst.append("-")

        pdbname = pdbpaths[modelnumlst.index(m)]
        pdbname_lst.append(pdbname)
                
    if not existuniprot:
        uniprotA_lst = ["-"] * len(proteinA_lst)
        uniprotB_lst = ["-"] * len(proteinB_lst)
            
    #print(len(proteinA_lst), len(proteinB_lst), len(iptmhighscore_lst), len(ptmhighscore_lst), len(pdbname_lst), len(paepng_lst), len(paejson_lst))

    df = pd.DataFrame(
        {'Protein A': proteinA_lst,
         'Protein B': proteinB_lst,
         'iptm': iptmhighscore_lst,
         'ptm': ptmhighscore_lst,
         'minPAE': mincrosspae_lst,
         'scaledPAE': inv_mincrosspae_lst,
         'Model': pdbname_lst,
         'PAE-png': paepng_lst,
         'PAE-json': paejson_lst,
         'SWISS-PROT Accessions Interactor A' : uniprotA_lst,
         'SWISS-PROT Accessions Interactor B' : uniprotB_lst
        })
        
    df.sort_values(by=[rankby], ascending=False, ignore_index=True, inplace=True)

    if skipped > 0:
        print("\n>> Warning: skipped " + str(skipped) + " since they were missing necessary files.")

    return(df)

def make_paeplot(pae, protnames, protlens):
    
    plt.imshow(pae, vmin=0, vmax=30, cmap="bwr")
    currlen=0
    yticklocs = []
    newprotnames = []
    for i, p in enumerate(protnames):
        newprotnames.append(p.split("_")[0]+"\n"+p.split("_")[2]+"-"+p.split("_")[3])
        
    for n, l in zip(protnames, protlens):
        
        yticklocs.append(int(l/2)+currlen)
        
        currlen = currlen + l
        
        plt.axvline(x=currlen-1, color='k')
        plt.axhline(y=currlen-1, color='k')
    
    plt.xlim(1, currlen-1)
    plt.ylim(1, currlen-1)
    plt.yticks(yticklocs, newprotnames, rotation='horizontal')
    plt.gca().invert_yaxis()
    
    return(plt)
    

def getinfo_frompaename(paejson):

    scores = json.loads(Path(paejson).read_text())

    pae = scores["pae"]            

    nameA = str(paejson).split("results/")[1].split("/")[0].split("-")[0]
    nameB = str(paejson).split("results/")[1].split("/")[0].split("-")[1]
    fasta=str(paejson).split("results/")[0]+"fastas/"+nameA+"-"+nameB+".fasta"
    with open(fasta, 'r') as f:
        lines = f.readlines()
    protnames = []
    protlens = []
    for i, a in enumerate(lines):
        if a[0]==">":
            protnames.append(a[1:])
            protlens.append(len(lines[i+1]))
    #lenprotA = int(nameA.split("_")[-1]) - int(nameA.split("_")[-2]) + 1
    #lenprotB = int(nameB.split("_")[-1]) - int(nameB.split("_")[-2]) + 1
    #nameprotA = nameA.split("_")[0] + " " + nameA.split("_")[1] + "\n" + nameA.split("_")[2]+ "-" + nameA.split("_")[3]
    #nameprotB = nameB.split("_")[0] + " " + nameB.split("_")[1] + "\n" + nameB.split("_")[2]+ "-" + nameB.split("_")[3]
    
    return(pae, protnames, protlens)

def getinfo_frompaename_legacy(paejson):

    scores = json.loads(Path(paejson).read_text())

    pae = scores[0]["distance"]
    splitby = int(math.sqrt(len(pae)))
    pae = [pae[i:i + splitby] for i in range(0, len(pae), splitby)]

    nameA = str(paejson).split("results/")[1].split("/")[0].split("-")[0]
    nameB = str(paejson).split("results/")[1].split("/")[0].split("-")[1]
    fasta=str(paejson).split("results/")[0]+"fastas/"+nameA+"-"+nameB+".fasta"
    with open(fasta, 'r') as f:
        lines = f.readlines()
    protnames = []
    protlens = []
    for i, a in enumerate(lines):
        if a[0]==">":
            protnames.append(a[1:])
            protlens.append(len(lines[i+1]))

    #lenprotA = int(nameA.split("_")[-1]) - int(nameA.split("_")[-2]) + 1
    #lenprotB = int(nameB.split("_")[-1]) - int(nameB.split("_")[-2]) + 1
    #nameprotA = nameA.split("_")[0] + " " + nameA.split("_")[1] + "\n" + nameA.split("_")[2]+ "-" + nameA.split("_")[3]
    #nameprotB = nameB.split("_")[0] + " " + nameB.split("_")[1] + "\n" + nameB.split("_")[2]+ "-" + nameB.split("_")[3]
    
    return(pae, protnames, protlens)

# def showpae(paejson):

#     pae, protnames, protlens = getinfo_frompaename(paejson)
#     plt = make_paeplot(pae, protnames, protlens)
#     plt.show()
#     return(len(protnames))

def getpae(paejson):
    try:
        return(getinfo_frompaename(paejson))
    except:
        try:
            return(getinfo_frompaename_legacy(paejson))
        except:
            sys.exit("\n>> Could not interpret the PAE json file.\n")

def getmincrosspae(paejson):

    pae, protnames, protlens = getpae(paejson)

    if len(protnames) == 2:

        subgrid = np.array(pae)[(protlens[0]+2):, 0:(protlens[0]-2)]

    elif len(protnames) == 3:

        if protnames[0] == protnames[1]:
            dimerlen = protlens[0]+protlens[1] # which should be the same
            subgrid = np.array(pae)[(dimerlen+2):, 0:(dimerlen-2)]

        else:
            dimerlen = protlens[1]+protlens[2] # which should be the same
            subgrid = np.array(pae)[(protlens[0]+2):, 0:(protlens[0]-2)]

    elif len(protnames) == 4:

        dimerlen = protlens[0]+protlens[1]
        subgrid = np.array(pae)[(dimerlen+2):, 0:(dimerlen-2)]

    return(subgrid.min())

def summarize_pae_pdf(df, threshold, rankby):
    
    if threshold == 0:
        pdfname = "PAEs-rankedby-" + rankby + ".pdf"
    else:
        pdfname = "PAEs-"+rankby+"-above-" + str(threshold).replace(".", "p") + ".pdf"
    
    print("\n>> Writing " + pdfname)

    with PdfPages(pdfname) as pdf:
        for i, result in df[df[rankby]>threshold].iterrows():
            pae, protnames, protlens = getpae(result["PAE-json"])
            plt = make_paeplot(pae, protnames, protlens)
            plt.title("\n".join(wrap(result["Model"], 80)), fontsize=8)
            pdf.savefig()
            plt.close('all')

# def show_pdb(modelname, chains, show_sidechains, show_mainchains=False, color="chain"):
    
#     view = py3Dmol.view(js='https://3dmol.org/build/3Dmol.js',)
#     view.addModel(open(modelname,'r').read(),'pdb')

#     if color == "lDDT":
#         view.setStyle({'cartoon': {'colorscheme': {'prop':'b','gradient': 'roygb','min':50,'max':90}}})
#     elif color == "rainbow":
#         view.setStyle({'cartoon': {'color':'spectrum'}})
#     elif color == "chain":
#         #chains = len(queries[0][1]) + 1 if is_complex else 1
#         for n,chain,color in zip(range(chains),list("BCDEFGH"),
#                          ["cyan", "magenta", "orange", "yellow"]): #"["lime","cyan","magenta","yellow","salmon","white","blue","orange"]"
#             view.setStyle({'chain':chain},{'cartoon': {'color':color}})
#     if show_sidechains:
#         BB = ['C','O','N']
#         view.addStyle({'and':[{'resn':["GLY","PRO"],'invert':True},{'atom':BB,'invert':True}]},
#                             {'stick':{'colorscheme':f"WhiteCarbon",'radius':0.3}})
#         view.addStyle({'and':[{'resn':"GLY"},{'atom':'CA'}]},
#                             {'sphere':{'colorscheme':f"WhiteCarbon",'radius':0.3}})
#         view.addStyle({'and':[{'resn':"PRO"},{'atom':['C','O'],'invert':True}]},
#                             {'stick':{'colorscheme':f"WhiteCarbon",'radius':0.3}})  
#     if show_mainchains:
#         BB = ['C','O','N','CA']
#         view.addStyle({'atom':BB},{'stick':{'colorscheme':f"WhiteCarbon",'radius':0.3}})
        
#     nameA = str(modelname).split("results/")[1].split("/")[0].split("-")[0]
#     nameB = str(modelname).split("results/")[1].split("/")[0].split("-")[1]
#     lenprotA = int(nameA.split("_")[-1]) - int(nameA.split("_")[-2]) + 1
#     lenprotB = int(nameB.split("_")[-1]) - int(nameB.split("_")[-2]) + 1
#     nameprotA = nameA.split("_")[0] + " " + nameA.split("_")[1] + "\n" + nameA.split("_")[2]+ "-" + nameA.split("_")[3]
#     nameprotB = nameB.split("_")[0] + " " + nameB.split("_")[1] + "\n" + nameB.split("_")[2]+ "-" + nameB.split("_")[3]
#     #view.addLabel(nameprotA,{'fontOpacity':1, 'fontSize':12, 'fontColor':'black','backgroundOpacity':0.2, 'backgroundColor':'cyan'},{'resi':1})
#     #view.addLabel(nameprotB,{'fontOpacity':1, 'fontSize':12, 'fontColor':'black','backgroundOpacity':0.2, 'backgroundColor':'magenta'},{'resi':lenprotA+10})

#     view.zoomTo()
#     return view

def write_top(df, threshold, rankby):

    if threshold == 0:
        basename = "Results"
    else:
        basename = "Results-"+rankby+"-above-" + str(threshold).replace(".", "p")
        
    excelname = basename + ".xlsx"
    csvname = basename + ".csv"

    print("\n>> Writing " + excelname + " and " + csvname)
    
    with pd.ExcelWriter(excelname) as writer:  
        df[df[rankby]>threshold].to_excel(writer)
        
    df.to_csv(csvname)
        
        
def write_modelpngs(df, threshold, rankby, overwrite=False):

    print("\n>> Writing model snapshots...")

    pymol.finish_launching(['pymol', '-qc']) #-Q will suppress render outputs too if you want

    total=0

    for i,result in df[df[rankby]>threshold].iterrows():
        
        model=result["Model"]
        png1 = model[:-4]+".png"
        png2 = model[:-4]+"-rotated.png"
        
        if not overwrite and os.path.exists(png1) and os.path.exists(png2):
            continue

        if os.path.exists(png1):
            os.rename(png1, model[:-4]+"-backup.png")
        
        cmd.load(model, "current")
        cmd.cartoon("automatic")
        cmd.bg_color("white")
        #cmd.ray(600,600)
        #cmd.draw(300,300,antialias=2)
        cmd.zoom()
        cmd.util.cbc()
        cmd.png(png1, width=600, height=600, dpi=900)

        waitfor(png1)

        if not os.path.exists(png1):
            if os.path.exists(model[:-4]+"-backup.png"):
                os.rename(model[:-4]+"-backup.png", png1)
            sys.exit("\n>> Error: pymol didn't output anything.\nFailed on: "+model+"\n")
        elif os.path.exists(model[:-4]+"-backup.png"):
            os.remove(model[:-4]+"-backup.png")
        
        cmd.rotate(axis='x',angle=90)
        cmd.rotate(axis='y',angle=90)
        cmd.zoom()
        cmd.png(png2, width=600, height=600, dpi=900)

        waitfor(png2)

        cmd.delete("current")
        
        total+=1
        
def waitfor(filename):
    pngexists=False
    tic = time.perf_counter()
    while not pngexists:
        if os.path.exists(filename):
            pngexists=True
        time.sleep(0.25)
        toc = time.perf_counter()
        if toc-tic > 4:
            sys.exit("\n>> Error: pymol didn't output anything in 4 seconds for file:\n"+filename)
    
def summarize_paeandmodel_pdf(df, threshold, rankby):
    
    if threshold == 0:
        pdfname = "PAEs-Models.pdf"
    else:
        pdfname = "PAEs-Models-"+rankby+"-above-"+ str(threshold).replace(".", "p") + ".pdf"
    
    print("\n>> Writing " + pdfname + "\n")

    with PdfPages(pdfname) as pdf:
        for i, result in df[df[rankby]>threshold].iterrows():
            
            if result["iptm"] == "-":
                iptmstring = "-"
                ptmstring = "-"
            else:
                iptmstring = str("{:.2f}".format(result["iptm"]))
                ptmstring = str("{:.2f}".format(result["ptm"]))
            
            if rankby=="scaledPAE":
                plottitle = [result["Model"],"iptm: "+ iptmstring + ", ptm: "+ ptmstring,"minimum-pae: "+str("{:.2f}".format(result["minPAE"])) + ", scaled-pae: "+str("{:.2f}".format(result["scaledPAE"])), result["Protein A"]+" ("+result["SWISS-PROT Accessions Interactor A"]+")",result["Protein B"]+" ("+result["SWISS-PROT Accessions Interactor B"]+")"]
            else:
                plottitle = [result["Model"],"iptm: "+ iptmstring + ", ptm: "+ iptmstring, result["Protein A"]+" ("+result["SWISS-PROT Accessions Interactor A"]+")",result["Protein B"]+" ("+result["SWISS-PROT Accessions Interactor B"]+")"]
            plottitle = "\n".join(plottitle)
            
            try:
                pae, protnames, protlens = getinfo_frompaename(result["PAE-json"])
            except:
                try:
                    pae, protnames, protlens = getinfo_frompaename_legacy(result["PAE-json"])
                except:
                    sys.exit("\n>> Could not interpret the PAE json file.\n")

            fig = plt.figure(figsize=(60,20))
            fig.suptitle(plottitle, fontsize=30)
            plt.subplot(1, 3, 1)
            make_paeplot(pae, protnames, protlens)
            plt.xticks(fontsize=36)
            plt.yticks(fontsize=36)

            png1 = result["Model"][:-4]+".png"
            png2 = result["Model"][:-4]+"-rotated.png"
            
            if not os.path.exists(png1):
                sys.exit("\n>> Error: The model snapshots could not be found.\n")
            elif not os.path.exists(png2):
                sys.exit("\n>> Error: The model snapshots could not be found.\n")
            
            plt.subplot(1, 3, 2)
            img1=plt.imread(png1)
            plt.axis('off')
            #img1=cv.imread(png1)
            plt.imshow(img1)
            
            plt.subplot(1, 3, 3)
            img2=plt.imread(png2)
            plt.axis('off')
            plt.imshow(img2)
            
            pdf.savefig()
            plt.close('all')

    print(">> If there is an error below, ignore it." + "\n")
